using System;
using System.Runtime.InteropServices;
using Yeti.WMFSdk;

namespace Yeti.VFW
{
  [StructLayout(LayoutKind.Sequential, CharSet=CharSet.Unicode)]
  public struct AVISTREAMINFO 
  {
    public uint		fccType;
    public uint    fccHandler;
    public uint    dwFlags;        /* Contains AVITF_* flags */
    public uint		dwCaps;
    public ushort	wPriority;
    public ushort	wLanguage;
    public uint    dwScale;
    public uint    dwRate; /* dwRate / dwScale == samples/second */
    public uint    dwStart;
    public uint    dwLength; /* In units above... */
    public uint		dwInitialFrames;
    public uint    dwSuggestedBufferSize;
    public uint    dwQuality;
    public uint    dwSampleSize;
    public RECT    rcFrame;
    public uint		dwEditCount;
    public uint		dwFormatChangeCount;
    [MarshalAs(UnmanagedType.ByValTStr, SizeConst=64)]
    public string	szName;
  }

  /// <summary>
	/// AVIFile functions wrapper
	/// </summary>
	internal sealed class AVIFileFuncts
	{
		private AVIFileFuncts() {	 }

    [DllImport("avifil32.dll")]
    public static extern void AVIFileInit();
    [DllImport("avifil32.dll")]
    public static extern void AVIFileExit();
    
    [DllImport("avifil32.dll", CharSet=CharSet.Unicode, ExactSpelling=false)]
    public static extern int AVIFileOpen (/*PAVIFILE*/out IntPtr ppfile, 
                                          string szFile,
                                          uint uMode, IntPtr lpHandler);
    [DllImport("avifil32.dll")]
    public static extern uint AVIFileRelease(/*PAVIFILE*/IntPtr pfile);
    [DllImport("avifil32.dll")]
    public static extern int AVIFileGetStream(/*PAVIFILE*/IntPtr pfile, 
                                              /*PAVISTREAM*/out IntPtr ppavi, 
                                              uint fccType, int lParam);
    [DllImport("avifil32.dll")]
    public static extern uint AVIStreamRelease(/*PAVISTREAM*/IntPtr pavi);
    [DllImport("avifil32.dll")]
    public static extern int AVIStreamStart(/*PAVISTREAM*/IntPtr pavi);
    [DllImport("avifil32.dll")]
    public static extern int AVIStreamLength(/*PAVISTREAM*/IntPtr pavi);
    public static int AVIStreamEnd(IntPtr pavi)
    {
      return AVIStreamStart(pavi)+AVIStreamLength(pavi);
    }
    [DllImport("avifil32.dll")]
    public static extern int AVIStreamReadFormat(/*PAVISTREAM*/IntPtr pavi, 
                                                 int lPos,
                                                 IntPtr lpFormat,
                                                 ref int lpcbFormat);
    [DllImport("avifil32.dll")]
    public static extern int AVIStreamReadFormat(/*PAVISTREAM*/IntPtr pavi, 
      int lPos,
      [Out]byte[] lpFormat,
      ref int lpcbFormat);
    public static int AVIStreamReadFormatSize(/*PAVISTREAM*/IntPtr pavi, int lPos, out int lpcbFormat)
    {
      lpcbFormat = 0;
      return AVIStreamReadFormat(pavi, lPos, IntPtr.Zero, ref lpcbFormat);
    }
    
    [DllImport("avifil32.dll", CharSet=CharSet.Unicode, ExactSpelling=false)]
    public static extern int AVIStreamInfo(/*PAVISTREAM*/IntPtr pavi, 
                                           out AVISTREAMINFO psi, int lSize);
    
    public const int AVISTREAMREAD_CONVENIENT = -1;
    [DllImport("avifil32.dll")]
    public static extern int AVIStreamRead(/*PAVISTREAM*/IntPtr pavi,
                                           int lStart,
                                           int lSamples,
                                           IntPtr lpBuffer,
                                           int cbBuffer,
                                           out int plBytes,
                                           out int plSamples);
    
	}

  public struct FOURCC
  {
    private uint m_Value;
    public FOURCC(byte b0, byte b1, byte b2, byte b3)
    {
      m_Value	=	((uint)b0 | ( (uint)(b1) << 8) | ((uint)(b2) << 16) | ((uint)(b3) << 24 ));
    }
    public FOURCC(char ch0, char ch1, char ch2, char ch3)
      :this((byte)ch0, (byte)ch1, (byte)ch2, (byte)ch3)
    {
    }
    public uint Value
    {
      get { return m_Value; }
      set { m_Value = value; }
    }

    public static implicit operator uint (FOURCC fcc)
    {
      return fcc.m_Value;
    }

    public static explicit operator FOURCC (uint val)
    {
      FOURCC res;
      res.m_Value = val;
      return res;
    }
  }
  public class AVIFileReader : IDisposable
  {
    private bool disposed = false;
    private IntPtr m_AviFile = IntPtr.Zero;
    private IntPtr m_AudioStream = IntPtr.Zero;
    private byte[] m_AudioFormat = null;
    private AVISTREAMINFO m_AudioStreamInfo;
    private int m_AudioSamples = -1;
    private int m_AudioCurrSample = -1;
    private IntPtr m_VideoStream = IntPtr.Zero;
    private byte[] m_VideoFormat = null;
    private AVISTREAMINFO m_VideoStreamInfo;
    private int m_VideoSamples = -1;
    private int m_VideoCurrSample = -1;

    public const int AVISTREAMREAD_CONVENIENT = -1; 

    public AVIFileReader(string filename)
    {
      AVIFileFuncts.AVIFileInit();
      try
      {
        Marshal.ThrowExceptionForHR(AVIFileFuncts.AVIFileOpen(out m_AviFile, filename, /*OF_SHARE_DENY_NONE*/0x00000040, IntPtr.Zero));
        if( AVIFileFuncts.AVIFileGetStream(m_AviFile, out m_AudioStream, new FOURCC('a', 'u', 'd', 's'), 0) >= 0 )
        {
          m_AudioSamples = AVIFileFuncts.AVIStreamEnd(m_AudioStream);
          m_AudioCurrSample = AVIFileFuncts.AVIStreamStart(m_AudioStream);
          if ( m_AudioSamples > m_AudioCurrSample )
          {
            int FormatSize;
            Marshal.ThrowExceptionForHR(AVIFileFuncts.AVIStreamReadFormatSize(m_AudioStream, 0, out FormatSize));
            m_AudioFormat = new byte[FormatSize];
            Marshal.ThrowExceptionForHR(AVIFileFuncts.AVIStreamReadFormat(m_AudioStream, 0, m_AudioFormat, ref FormatSize));
            Marshal.ThrowExceptionForHR(AVIFileFuncts.AVIStreamInfo(m_AudioStream, out m_AudioStreamInfo, Marshal.SizeOf(m_AudioStreamInfo)));
          }
          else
          {
            AVIFileFuncts.AVIStreamRelease(m_AudioStream);
            m_AudioStream = IntPtr.Zero;
          }
        }
        
        if( AVIFileFuncts.AVIFileGetStream(m_AviFile, out m_VideoStream, new FOURCC('v', 'i', 'd', 's'), 0) >= 0 )
        {
          m_VideoSamples = AVIFileFuncts.AVIStreamEnd(m_VideoStream);
          m_VideoCurrSample = AVIFileFuncts.AVIStreamStart(m_VideoStream);
          if ( m_VideoSamples > m_VideoCurrSample )
          {
            int FormatSize;
            Marshal.ThrowExceptionForHR(AVIFileFuncts.AVIStreamReadFormatSize(m_VideoStream, 0, out FormatSize));
            m_VideoFormat = new byte[FormatSize];
            Marshal.ThrowExceptionForHR(AVIFileFuncts.AVIStreamReadFormat(m_VideoStream, 0, m_VideoFormat, ref FormatSize));
            Marshal.ThrowExceptionForHR(AVIFileFuncts.AVIStreamInfo(m_VideoStream, out m_VideoStreamInfo, Marshal.SizeOf(m_VideoStreamInfo)));
          }
          else
          {
            AVIFileFuncts.AVIStreamRelease(m_VideoStream);
            m_VideoStream = IntPtr.Zero;
          }
        }
      }
      catch
      {
        Dispose(false);
        throw;
      }
    }

    ~AVIFileReader()      
    {
      Dispose(false);
    }

    private void Dispose(bool disposing)
    {
      if(!this.disposed)
      {
        if(disposing)
        {
          // Dispose managed resources.
        }
             
        if ( m_VideoStream != IntPtr.Zero )
        {
          AVIFileFuncts.AVIStreamRelease(m_VideoStream);
          m_VideoStream = IntPtr.Zero;
        }

        if ( m_AudioStream != IntPtr.Zero )
        {
          AVIFileFuncts.AVIStreamRelease(m_AudioStream);
          m_AudioStream = IntPtr.Zero;
        }

        if ( m_AviFile != IntPtr.Zero )
        {
          AVIFileFuncts.AVIFileRelease(m_AviFile);
          m_AviFile = IntPtr.Zero;
        }
        AVIFileFuncts.AVIFileExit();
      }
      disposed = true;         
    }

    #region IDisposable Members

    public void Dispose()
    {
      Dispose(true);
      GC.SuppressFinalize(this);
    }

    #endregion

    public bool HasAudio
    {
      get { return m_AudioStream != IntPtr.Zero; }
    }

    public bool HasVideo
    {
      get { return m_VideoStream != IntPtr.Zero; }
    }

    public byte[] AudioFormat
    {
      get { return m_AudioFormat; }
    }

    public byte[] VideoFormat
    {
      get { return m_VideoFormat; }
    }

    public AVISTREAMINFO AudioStreamInfo
    {
      get { return m_AudioStreamInfo; }
    }

    public AVISTREAMINFO VideoStreamInfo
    {
      get { return m_VideoStreamInfo; }
    }

    public uint AudioSamples
    {
      get { return (m_AudioSamples > 0)? (uint)m_AudioSamples:0; }
    }

    public uint VideoSamples
    {
      get { return (m_VideoSamples > 0)? (uint)m_VideoSamples:0; }
    }

    //Audio read samples functions
    public void GetAudioSamplesSizes(int Start, int DesiredSamples, out int NeededBufferSize, out int RealSamplesToRead)
    {
      if (HasAudio)
      {
        Marshal.ThrowExceptionForHR(AVIFileFuncts.AVIStreamRead(m_AudioStream, Start, DesiredSamples, IntPtr.Zero, 0, out NeededBufferSize, out RealSamplesToRead));
      }
      else
      {
        throw new InvalidOperationException("There is no audio stream in this AVI file");
      }
    }

    public void GetAudioSamplesSizes(int Start, out int NeededBufferSize, out int RealSamplesToRead)
    {
      GetAudioSamplesSizes(Start, AVISTREAMREAD_CONVENIENT, out NeededBufferSize, out RealSamplesToRead);
    }

    public void ReadAudioSamples(int Start, int DesiredSamples, IntPtr Buff, int BufferSize, out int BytesRead, out int SamplesRead)
    {
      if (HasAudio)
      {
        Marshal.ThrowExceptionForHR(AVIFileFuncts.AVIStreamRead(m_AudioStream, Start, DesiredSamples, Buff, BufferSize, out BytesRead, out SamplesRead));
      }
      else
      {
        throw new InvalidOperationException("There is no audio stream in this AVI file");
      }
    }

    public byte[] GetAudioSamples(int Start, int DesiredSamples, out int SamplesRead)
    {
      byte[] result;
      int BufferSize;
      GetAudioSamplesSizes(Start, DesiredSamples, out BufferSize, out SamplesRead);
      result = new byte[BufferSize];
      GCHandle h = GCHandle.Alloc(result, GCHandleType.Pinned);
      try
      {
        ReadAudioSamples(Start, DesiredSamples, h.AddrOfPinnedObject(), BufferSize, out BufferSize, out SamplesRead);
      }
      finally
      {
        h.Free();
      }
      return result;
    }

    public byte[] GetAudioSamples(int Start, out int SamplesRead)
    {
      return GetAudioSamples(Start, AVISTREAMREAD_CONVENIENT, out SamplesRead);
    } 

    //Video read samples functions
    public void GetVideoSamplesSizes(int Start, int DesiredSamples, out int NeededBufferSize, out int RealSamplesToRead)
    {
      if (HasVideo)
      {
        Marshal.ThrowExceptionForHR(AVIFileFuncts.AVIStreamRead(m_VideoStream, Start, DesiredSamples, IntPtr.Zero, 0, out NeededBufferSize, out RealSamplesToRead));
      }
      else
      {
        throw new InvalidOperationException("There is no Video stream in this AVI file");
      }
    }

    public void GetVideoSamplesSizes(int Start, out int NeededBufferSize, out int RealSamplesToRead)
    {
      GetVideoSamplesSizes(Start, AVISTREAMREAD_CONVENIENT, out NeededBufferSize, out RealSamplesToRead);
    }

    public void ReadVideoSamples(int Start, int DesiredSamples, IntPtr Buff, int BufferSize, out int BytesRead, out int SamplesRead)
    {
      if (HasVideo)
      {
        Marshal.ThrowExceptionForHR(AVIFileFuncts.AVIStreamRead(m_VideoStream, Start, DesiredSamples, Buff, BufferSize, out BytesRead, out SamplesRead));
      }
      else
      {
        throw new InvalidOperationException("There is no Video stream in this AVI file");
      }
    }

    public byte[] GetVideoSamples(int Start, int DesiredSamples, out int SamplesRead)
    {
      byte[] result;
      int BufferSize;
      GetVideoSamplesSizes(Start, DesiredSamples, out BufferSize, out SamplesRead);
      result = new byte[BufferSize];
      GCHandle h = GCHandle.Alloc(result, GCHandleType.Pinned);
      try
      {
        ReadVideoSamples(Start, DesiredSamples, h.AddrOfPinnedObject(), BufferSize, out BufferSize, out SamplesRead);
      }
      finally
      {
        h.Free();
      }
      return result;
    }

    public byte[] GetVideoSamples(int Start, out int SamplesRead)
    {
      return GetVideoSamples(Start, AVISTREAMREAD_CONVENIENT, out SamplesRead);
    }
  }
}
