using System;
using System.Runtime.InteropServices;
using Yeti.MMedia;
using Yeti.WMFSdk;
using Yeti.VFW;

namespace Avi2Wmv
{ 
  public class ProgressEventArgs : EventArgs
  {
    internal ulong m_Time;
    internal int m_Percent;
    private bool m_Cancel = false;
    internal ProgressEventArgs(int percent, ulong time)
    {
      m_Percent = percent;
      m_Time = time;
    }
    public int Percent
    {
      get { return m_Percent; }
    }
    public ulong Time
    {
      get { return m_Time; }
    }
    public bool Cancel
    {
      get { return m_Cancel; }
      set { m_Cancel = value; }
    }
  }

  public delegate void ProgressEventHandler(object Sender, ProgressEventArgs e); 

  /// <summary>
	/// Convert an uncrompressed AVI file to a Windows Media Video file.
	/// </summary>
	public class AviToAsf : IDisposable
	{
		AVIFileReader m_AviFile = null;
    IWMWriter m_Writer = null;
    uint m_AudioSamples = 0;
    uint m_AudioInput;
    uint m_VideoSamples = 0;
    uint m_VideoInput;
    
    public AviToAsf(string InAviFilename, string OutAsfFilename, IWMProfile Profile, WM_Attr[] MetadataAttributes)
		{
			m_AviFile = new AVIFileReader(InAviFilename);
      try
      {
        m_Writer = WM.CreateWriter();
        m_Writer.SetOutputFilename(OutAsfFilename);
        InitInputs(Profile);
        if ( MetadataAttributes != null )
        {
          WMHeaderInfo info = new WMHeaderInfo((IWMHeaderInfo)m_Writer);
          foreach(WM_Attr attr in MetadataAttributes)
          {
            info.SetAttribute(attr);
          }
          info = null;
        }
      }
      catch
      {
        m_AviFile.Dispose();
        throw;
      }
    }

    public event ProgressEventHandler Progress;

    private void DoProgress(ProgressEventArgs e)
    {
      if (Progress != null)
      {
        Progress(this, e);
      }
    }

    public bool Start()
    {
      ulong AudioPresentTime = 0;
      ulong VideoPresentTime = 0;
      ulong TotalSamples = m_AudioSamples + m_VideoSamples;
      ulong CurrentSample = 0;
      uint CurrentAudioSample = 0;
      uint CurrentVideoSample = 0;

      ProgressEventArgs ProgArgs = new ProgressEventArgs(0, 0);
      m_Writer.BeginWriting();
      DoProgress(ProgArgs);
      while ( (CurrentSample < TotalSamples ) && (!ProgArgs.Cancel) )
      {
        if ( (CurrentAudioSample < m_AudioSamples ) &&//There are audio samples and their presentation time is before of the video samples.
             !((CurrentVideoSample < m_VideoSamples) && (AudioPresentTime > VideoPresentTime)) )
        { //Process audio samples
          int Samples;
          int BufferSize;
          INSSBuffer buffer;
          IntPtr BufferPtr;
          m_AviFile.GetAudioSamplesSizes((int)CurrentAudioSample, out BufferSize, out Samples);
          m_Writer.AllocateSample((uint)BufferSize, out buffer);
          buffer.GetBuffer(out BufferPtr);
          m_AviFile.ReadAudioSamples((int)CurrentAudioSample, Samples, BufferPtr, BufferSize, out BufferSize, out Samples);
          buffer.SetLength((uint)BufferSize);
          m_Writer.WriteSample(m_AudioInput, AudioPresentTime, 0, buffer);
          CurrentAudioSample += (uint)Samples;
          AudioPresentTime = 10000000 * (ulong)CurrentAudioSample * m_AviFile.AudioStreamInfo.dwScale / m_AviFile.AudioStreamInfo.dwRate;
        }
        if ( (CurrentVideoSample < m_VideoSamples ) &&
             !((CurrentAudioSample < m_AudioSamples) && (AudioPresentTime < VideoPresentTime)) )
        {//Process video samples
          int Samples;
          int BufferSize;
          
          m_AviFile.GetVideoSamplesSizes((int)CurrentVideoSample, out BufferSize, out Samples);
          if ( BufferSize > 0)
          {
            INSSBuffer buffer;
            IntPtr BufferPtr;
            m_Writer.AllocateSample((uint)BufferSize, out buffer);
            buffer.GetBuffer(out BufferPtr);
            m_AviFile.ReadVideoSamples((int)CurrentVideoSample, Samples, BufferPtr, BufferSize, out BufferSize, out Samples);
            buffer.SetLength((uint)BufferSize);
            m_Writer.WriteSample(m_VideoInput, VideoPresentTime, 0, buffer);
          }
          else
          { // 0-zized sample, just skip the sample
            Samples = 1;
          }
          CurrentVideoSample += (uint)Samples;
          VideoPresentTime = 10000000 * (ulong)CurrentVideoSample * m_AviFile.VideoStreamInfo.dwScale / m_AviFile.VideoStreamInfo.dwRate;
        }
        CurrentSample = CurrentAudioSample + CurrentVideoSample;
        ProgArgs.m_Percent = (int)(CurrentSample * 100 / TotalSamples);
        ProgArgs.m_Time = Math.Max(AudioPresentTime, VideoPresentTime);
        DoProgress(ProgArgs);
      }
      m_Writer.EndWriting();
      m_Writer = null;
      return !ProgArgs.Cancel;
    }

    private static WM_MEDIA_TYPE WMTypeFromAviAudioStream(AVIFileReader AviFile)
    {
      if ( AviFile.HasAudio )
      {
        WM_MEDIA_TYPE res;
        res.majortype = MediaTypes.WMMEDIATYPE_Audio;
        res.subtype = MediaTypes.WMMEDIASUBTYPE_PCM;
        res.bFixedSizeSamples = true;
        res.bTemporalCompression = false;
        res.lSampleSize = AviFile.AudioStreamInfo.dwSampleSize;
        res.formattype = MediaTypes.WMFORMAT_WaveFormatEx;
        res.pUnk = IntPtr.Zero;
        res.cbFormat = (uint)AviFile.AudioFormat.Length;
        //Must be free using Marshal.FreeCoTaskMem
        res.pbFormat = Marshal.AllocCoTaskMem((int)res.cbFormat);
        try
        {
          Marshal.Copy(AviFile.AudioFormat, 0, res.pbFormat, (int)res.cbFormat);
          return res;
        }
        catch
        {
          Marshal.FreeCoTaskMem(res.pbFormat);
          throw;
        }
      }
      else
      {
        throw new ArgumentException("There is no audio stream in the AVI file"); 
      }
    }
    
    private static WM_MEDIA_TYPE WMTypeFromAviVideoStream(AVIFileReader AviFile)
    {
      if ( AviFile.HasVideo )
      {
        WM_MEDIA_TYPE res;
        res.majortype = MediaTypes.WMMEDIATYPE_Video;
        
        res.cbFormat = (uint)(Marshal.SizeOf(typeof(WMVIDEOINFOHEADER)) - Marshal.SizeOf(typeof(BITMAPINFOHEADER)) + AviFile.VideoFormat.Length);
        //Must be free using Marshal.FreeCoTaskMem
        res.pbFormat = Marshal.AllocCoTaskMem((int)res.cbFormat);
        try
        {
          //Copy the BITMAPINFOHEADER format to WMVIDEOINFOHEADER pointer
          const int bmiHeaderOffset = 48;//16 + 16 + 4 + 4 + 8
          Marshal.Copy(AviFile.VideoFormat, 0, (IntPtr)(res.pbFormat.ToInt32()+bmiHeaderOffset), AviFile.VideoFormat.Length);
          WMVIDEOINFOHEADER VideoInfo = (WMVIDEOINFOHEADER)Marshal.PtrToStructure(res.pbFormat, typeof(WMVIDEOINFOHEADER));
          VideoInfo.rcSource.left = 0;
          VideoInfo.rcSource.top = 0;
          VideoInfo.rcSource.bottom = VideoInfo.bmiHeader.biHeight;
          VideoInfo.rcSource.right = VideoInfo.bmiHeader.biWidth;
          VideoInfo.rcTarget = VideoInfo.rcSource;
          VideoInfo.dwBitRate = (uint)(((ulong)AviFile.VideoStreamInfo.dwSuggestedBufferSize* 8 * (ulong)AviFile.VideoStreamInfo.dwRate)/AviFile.VideoStreamInfo.dwScale);
          VideoInfo.dwBitErrorRate  = 0;
          VideoInfo.AvgTimePerFrame = (long)(10000000 * (ulong)AviFile.VideoStreamInfo.dwScale / AviFile.VideoStreamInfo.dwRate);
          //Update the WMVIDEOINFOHEADER pointer
          Marshal.StructureToPtr(VideoInfo, res.pbFormat, false);
        
          //
          // Map to the correct subtype GUID. If we don't map, just set it
          // to GUID_NULL; this probably means the SDK doesn't support the
          // input, but we'll still try to set the properties, just in case.
          //
          if( VideoInfo.bmiHeader.biCompression == BITMAPINFOHEADER.BI_RGB )
          {
            if( VideoInfo.bmiHeader.biBitCount == 32 )
            {
              res.subtype = MediaTypes.WMMEDIASUBTYPE_RGB32;
            }
            else if( VideoInfo.bmiHeader.biBitCount == 24 )
            {
              res.subtype = MediaTypes.WMMEDIASUBTYPE_RGB24;
            }
            else if( VideoInfo.bmiHeader.biBitCount == 16 )
            {
              res.subtype = MediaTypes.WMMEDIASUBTYPE_RGB555;
            }
            else if( VideoInfo.bmiHeader.biBitCount == 8 )
            {
              res.subtype = MediaTypes.WMMEDIASUBTYPE_RGB8;
            }
            else if( VideoInfo.bmiHeader.biBitCount == 4 )
            {
              res.subtype = MediaTypes.WMMEDIASUBTYPE_RGB4;
            }
            else if( VideoInfo.bmiHeader.biBitCount == 1 )
            {
              res.subtype = MediaTypes.WMMEDIASUBTYPE_RGB1;
            }
            else
            {
              res.subtype = Guid.Empty;
            }
          }
          else if( VideoInfo.bmiHeader.biCompression ==  new FOURCC('I', '4', '2', '0') )
          {
            res.subtype = MediaTypes.WMMEDIASUBTYPE_I420;
          }
          else if( VideoInfo.bmiHeader.biCompression == new FOURCC('I', 'Y','U', 'V') )
          {
            res.subtype = MediaTypes.WMMEDIASUBTYPE_IYUV;
          }
          else if( VideoInfo.bmiHeader.biCompression == new FOURCC('Y', 'V', '1', '2') )
          {
            res.subtype = MediaTypes.WMMEDIASUBTYPE_YV12;
          }
          else if( VideoInfo.bmiHeader.biCompression == new FOURCC('Y', 'U', 'Y', '2') )
          {
            res.subtype = MediaTypes.WMMEDIASUBTYPE_YUY2;
          }
          else if( VideoInfo.bmiHeader.biCompression == new FOURCC('U', 'Y', 'V', 'Y') )
          {
            res.subtype = MediaTypes.WMMEDIASUBTYPE_UYVY;
          }
          else if( VideoInfo.bmiHeader.biCompression == new FOURCC('Y', 'V', 'Y', 'U') )
          {
            res.subtype = MediaTypes.WMMEDIASUBTYPE_YVYU;
          }
          else if( VideoInfo.bmiHeader.biCompression == new FOURCC('Y', 'V', 'U', '9') )
          {
            res.subtype = MediaTypes.WMMEDIASUBTYPE_YVU9;
          }
          else
          {
            res.subtype = Guid.Empty;
          }

          res.bFixedSizeSamples = true;
          res.bTemporalCompression = false;
          res.lSampleSize = AviFile.VideoStreamInfo.dwSampleSize;
          res.formattype = MediaTypes.WMFORMAT_VideoInfo;
          res.pUnk = IntPtr.Zero;
          return res;
        }
        catch
        {
          Marshal.FreeCoTaskMem(res.pbFormat);
          throw;
        }
      }
      else
      {
        throw new ArgumentException("There is no audio stream in the AVI file"); 
      }
    }

    private void InitInputs(IWMProfile Profile)
    {
      uint cInputs;
      uint i = 0;
      ///* Uncomment the following code if you don't want modify the profile
        //In order to remove all streams with no matching inputs in the AVI file
      Profile.GetStreamCount(out cInputs);
      while (i < cInputs)
      { //Remove unused streams from profile
        IWMStreamConfig StrConfig;
        Guid mtype;
        Profile.GetStream(i, out StrConfig);
        StrConfig.GetStreamType(out mtype);
        if ( !(((mtype == MediaTypes.WMMEDIATYPE_Audio) && m_AviFile.HasAudio) ||
          ((mtype == MediaTypes.WMMEDIATYPE_Video) && m_AviFile.HasVideo)) )
        {
          Profile.RemoveStream(StrConfig);
          Profile.GetStreamCount(out cInputs);
        }
        else
        {
          i++;
        }
      }
      //*/
      m_Writer.SetProfile(Profile);
      m_Writer.GetInputCount(out cInputs);
      for (i = 0; i < cInputs; i++)
      {
        IWMInputMediaProps InProps = null;
        Guid mtype;
        m_Writer.GetInputProps(i, out InProps);
        InProps.GetType(out mtype);
        if ( mtype == MediaTypes.WMMEDIATYPE_Audio )
        {
          WM_MEDIA_TYPE InAudioWMTYpe = WMTypeFromAviAudioStream(m_AviFile);
          try
          {
            InProps.SetMediaType(ref InAudioWMTYpe);
            m_Writer.SetInputProps(i, InProps);
            m_AudioSamples = m_AviFile.AudioSamples;
            m_AudioInput = i;
          }
          finally
          {
            Marshal.FreeCoTaskMem(InAudioWMTYpe.pbFormat);
          }
        }
        else if ( mtype == MediaTypes.WMMEDIATYPE_Video ) 
        {
          WM_MEDIA_TYPE InVideoWMTYpe = WMTypeFromAviVideoStream(m_AviFile);
          try
          {
            InProps.SetMediaType(ref InVideoWMTYpe);
            m_Writer.SetInputProps(i, InProps);
            m_VideoSamples = m_AviFile.VideoSamples;
            m_VideoInput = i;
          }
          finally
          {
            Marshal.FreeCoTaskMem(InVideoWMTYpe.pbFormat);
          }
        }
        else
        {
          throw new ArgumentException("Unrecognized stream in the WM profile", "Profile");
        }
      }
    }

    #region IDisposable Members

    public void Dispose()
    {
      if ( m_Writer != null )
      {
        m_Writer.EndWriting();
        m_Writer = null;
      }
      if ( m_AviFile != null )
      {
        m_AviFile.Dispose();
        m_AviFile = null;
      }
      
    }

    #endregion
  }
}
